package com.jclark.xsl;

import com.jclark.xml.output.*;
import java.io.IOException;
import java.io.OutputStream;

public class ResultImpl implements Result {
  XMLWriter writer;
  NamespacePrefixMap currentMap = null;

  public ResultImpl(OutputStream outputStream) throws IOException {
    writer = new UTF8XMLWriter(outputStream,
			       UTF8XMLWriter.MINIMIZE_EMPTY_ELEMENTS);
  }

  public
  void characters(String str) throws XSLException {
    try {
      writer.write(str);
    }
    catch (IOException e) {
      throw new XSLException(e);
    }
  }

  public
  void element(Name elementType, AttributeList attributes, NamespacePrefixMap map,
               Action content, ProcessContext context, Node sourceNode) throws XSLException {
    try {
      NamespacePrefixMap oldMap = currentMap;
      markup();
      writer.startElement(elementType.toString());
      if (map != currentMap) {
	int size = map.getSize();
	boolean changed = false;
	for (int i = 0; i < size; i++) {
	  String prefix = map.getPrefix(i);
	  String namespace = map.getNamespace(i);
	  if (currentMap == null
	      || !namespace.equals(currentMap.getNamespace(prefix))) {
	    changed = true;
	    writer.attribute("xmlns:" + prefix, namespace);
	  }
	}
	String defaultNamespace = map.getDefaultNamespace();
	if (defaultNamespace == null) {
	  if (currentMap != null && currentMap.getDefaultNamespace() != null) {
	    writer.attribute("xmlns", "");
	    changed = true;
	  }
	}
	else if (currentMap == null
	         || !defaultNamespace.equals(currentMap.getDefaultNamespace())) {
	  changed = true;
	  writer.attribute("xmlns", defaultNamespace);
	}
	if (changed)
	  currentMap = map;
      }
      int nAtts = attributes.getAttributeCount();
      for (int i = 0; i < nAtts; i++) {
        Name name = attributes.getAttributeName(i);
	String nameString;
	String namespace = name.getNamespace();
	if (namespace == null)
	  nameString = name.toString();
	else {
	  String prefix = name.getPrefix();
	  if (namespace.equals(currentMap.getNamespace(prefix)))
	    nameString = name.toString();
	  else {
	    // This can happen with xsl:use.
	    String newPrefix = currentMap.getPrefix(namespace);
	    if (newPrefix == null && oldMap != currentMap && oldMap != null)
	      newPrefix = oldMap.getPrefix(namespace);
	    if (newPrefix != null)
	      nameString = newPrefix + ":" + name.getLocalPart();
	    else {
	      if (currentMap.getNamespace(prefix) == null) {
		currentMap = currentMap.bind(prefix, namespace);
		writer.attribute("xmlns:" + prefix, namespace);
		nameString = name.toString();
	      }
	      else {
		// Must generate our own prefix...
		for (int j = 0;; j++) {
		  newPrefix = "ns" + Integer.toString(j);
		  if (currentMap.getPrefix(newPrefix) == null)
		    break;
		}
		nameString = newPrefix + ":" + name.getLocalPart();
		currentMap = currentMap.bind(newPrefix, namespace);
		writer.attribute("xmlns:" + newPrefix, namespace);
	      }
	    }
	  }
	}
	writer.attribute(nameString,
			 attributes.getAttributeValue(i));
      }
      if (content != null) {
	content.invoke(context, sourceNode, this);
	markup();
      }
      writer.endElement(elementType.toString());
      currentMap = oldMap;
    }
    catch (IOException e) {
      throw new XSLException(e);
    }
  }

  protected void markup() throws XSLException { }

  public void finish() throws XSLException {
    try {
      writer.close();
    }
    catch (IOException e) {
      throw new XSLException(e);
    }
  }

}
